local vector = require 'vector'
local ffi = require 'ffi'
local C = ffi.C
ffi.cdef[[
typedef struct device device;
typedef struct sample sample;
typedef struct channel channel;
device* deviceOpen (const char* devname) ;
device* deviceGetCurrent (void) ;
bool deviceSetCurrent (device* dev) ;
void deviceClose (device* dev) ;
const char* deviceList (void) ;
sample* sampleLoad (const char* filename) ;
double sampleGetLength (sample* sample) ;
double sampleGetAttr (sample* sample, int attr) ;
void sampleFree (sample* sample) ;
channel* channelNew (sample* sample) ;
channel* streamLoad (const char* filename) ;
void channelPlay (channel* ch) ;
void channelPause (channel* ch) ;
void channelStop (channel* ch) ;
void channelWait (channel* ch, int timeout) ;
double channelGetLength (channel* ch) ;
int channelGetType (channel* ch) ;
//int channelGetFlags (channel* ch) ;
//void channelSetFlags (channel* ch, int flags) ;
sample* channelGetSample (channel* ch) ;
void channelSetSample (channel* ch, sample* sample) ;
double channelGetAttr (channel* ch, int attr) ;
void channelSetAttr (channel* ch, int attr, double value) ;
void channelGetAttrv (channel* ch, int attr, float* v) ;
void channelSetAttrv (channel* ch, int attr, const float* v) ;
void channelFree (channel* ch) ;
void channelGC (channel* ch) ;
double listenerGetAttr (int attr) ;
void listenerSetAttr (int attr, double value) ;
void listenerGetAttrv (int attr, float* v) ;
void listenerSetAttrv (int attr, const float* v) ;
]]

local device = {
open = C.deviceOpen,
close = C.deviceClose,
}

local sample = {
free = C.sampleFree,
}

local channel = {
new = C.channelNew,
pause = C.channelPause,
stop = C.channelStop,
wait = function (self, timeout) return C.channelWait(self, timeout or -1) end,
free = C.channelFree,
}

local listener = {
inverseDistance = 0xD001,
inverseDistanceClamped = 0xD002,
linearDistance = 0xD003,
linearDistanceClamped = 0xD004,
exponentDistance = 0xD005,
exponentDistanceClamped = 0xD006
}

local audio = {
sample=sample,
channel=channel,
device=device,
listener=listener
}

local channelAttrs = {
volume = 0x100A, gain = 0x100A,
pan = 0x1A01,
pitch = 0x1003,
looped = 0x1007,
offset = 0x1024,
state = 0x1010,
relative = 0x202,
autofree = 0x1A02,
referenceDistance = 0x1020,
maxDistance = 0x1023,
minGain = 0x100D,
maxGain = 0x100E,
coneInnerAngle = 0x1001,
coneOuterAngle = 0x1002,
coneOuterGain = 0x1022,
rolloffFactor = 0x1021,
direct = 0x1033
}

local channelAttrvs = {
position = 0x1004,
direction = 0x1005,
velocity = 0x1006,
}

local sampleAttrs = {
frequency = 0x2001, samplerate = 0x2001, sampleRate = 0x2001,
bits = 0x2002, resolution = 0x2002,
channels = 0x2003,
size = 0x2004
}

local listenerAttrs = {
volume = 0x100A,
gain = 0x100A,
dopplerFactor = 0xC000,
dopplerVelocity = 0xC001,
speedOfSound = 0xC003,
distanceModel = 0xD000
}

local listenerAttrvs = {
position = 0x1004,
direction = 0x1005,
velocity = 0x1006,
at = 0x1005,
up = 0x1A00,
}

local deviceAttrs = {
frequency = 0x1007, sampleRate = 0x1007,
refreshRate = 0x1008, monoSources = 0x1010, stereoSources = 0x1011
}

function audio.init (name) 
rawset(audio, 'currentDevice', device.open(name)) 
return audio.currentDevice
end

function audio.free () 
device.close(audio.currentDevice) 
end

function audio.loadStream (file)
local stream = C.streamLoad(file)
if stream then
ffi.gc(stream, C.channelGC)
return stream
end end

function audio.loadSample (file)
local sample = C.sampleLoad(file)
if sample then
ffi.gc(sample, C.sampleFree)
return sample
end end

function device.list ()
local s = C.deviceList()
local t = {}
local i = 1
while s[0]>0 do
local name = ffi.string(s)
t[i] = name
i = i + 1
s = s + #name +1
end
return t
end

function device.__index (self, name, val)
if deviceAttrs[name] then return C.listenerGetAttr(deviceAttrs[name])
else return device[name]
end end

function device.__newindex (self, name, val)
end

function sample.play (self, attrs)
local ch = channel.new(self)
if not ch then return end
ch.autofree = true
ffi.gc(ch, C.channelGC)
return ch:play(attrs)
end

function sample.__index (self, name)
if sampleAttrs[name] then return C.sampleGetAttr(self, sampleAttrs[name])
elseif name=='length' or name=='duration' then return C.sampleGetLength(self)
else return sample[name]
end end

function channel.play (self, attrs)
if attrs then
for k,v in pairs(attrs) do self[k]=v end
end
C.channelPlay(self)
return self
end

function channel.__index (self, name)
if channelAttrs[name] then return C.channelGetAttr(self, channelAttrs[name])
elseif channelAttrvs[name] then
local v = vector.new()
C.channelGetAttrv(self, channelAttrvs[name], v.v)
return v
elseif name=='length' or name=='duration' then return C.channelGetLength(self)
elseif name=='playing' then return self.state == 0x1012
elseif name == 'paused' then return self.state == 0x1013
elseif name == 'stopped' then return self.state == 0x1014
elseif name == 'semitones' then return 12 * math.log(self.pitch) / math.log(2)
elseif name=='sample' then return C.channelGetSample(self)
elseif name=='type' then
local type = C.channelGetType(self)
if type == 0x1028 then return 'sample'
elseif type == 0x1029 then return 'stream'
else return nil
end
elseif name == 'stream' then return self.type=='stream'
else return channel[name]
end end

function channel.__newindex (self, name, val)
if channelAttrs[name] then C.channelSetAttr(self, channelAttrs[name], val)
elseif channelAttrvs[name] then
if type(val)=='table' then val = vector.new(val.x or val[1] or 0, val.y or val[2] or 0, val.z or val[3] or 0) end
C.channelSetAttrv(self, channelAttrvs[name], val.v)
elseif name == 'semitones' then self.pitch = 2^(val/12)
elseif name=='sample' then C.channelSetSample(self, val)
elseif name == 'playing' and val and not self.playing then self:play()
elseif name == 'playing' and not val and self.playing then self:pause()
elseif name == 'paused' and val and not self.paused then self:pause()
elseif name == 'paused' and not val and self.paused then self:play()
elseif name == 'stopped' and val and not self.stopped then self:stop()
elseif name == 'stopped' and not val and self.stopped then self:play()
end end

function listener.__index (self, name)
if listenerAttrs[name] then return C.listenerGetAttr(listenerAttrs[name])
elseif listenerAttrvs[name] then
local v = vector.new()
C.listenerGetAttrv(listenerAttrvs[name], v.v)
return v
end end

function listener.__newindex (self, name, value)
if listenerAttrs[name] then C.listenerSetAttr(listenerAttrs[name], val)
elseif listenerAttrvs[name] then
if type(val)=='table' then val = vector.new(val.x or val[1] or 0, val.y or val[2] or 0, val.z or val[3] or 0) end
C.listenerSetAttrv(listenerAttrvs[name], val.v)
end end

function listener.position (x, y, z)
return lAttr3(0x1004, x,y,z)
end

setmetatable(listener, listener)
ffi.metatype('device',device)
ffi.metatype('sample',sample)
ffi.metatype('channel',channel)
return audio