#include "audio.h"

static device* curdev = NULL;

static ALuint getSourceFromBuffer (ALuint albuf) {
ALuint src = 0;
alGenSources(1,&src);
if (!src) return 0;
alSourcei(src, AL_BUFFER, albuf);
return src;
}

static ALuint loadSampleImpl (void* handle, audioIO* io, audioInfo* info) {
ALuint albuf = 0;
alGenBuffers(1,&albuf);
if (!albuf) { 
io->close(handle,NULL, NULL);
return 0; 
}
int sampleRate = info->samplerate, channels = info->channels, length = info->frames, buflen = info->frames*info->channels, n=0, m;
short* buffer = malloc(buflen*sizeof(short));
while (n<buflen) {
m = io->read(handle, FORMAT_S16, buffer+n, buflen-n);
if (m<0) break;
n+=m;
}
io->close(handle,NULL, NULL);
alBufferData(albuf, channels==2? AL_FORMAT_STEREO16 : AL_FORMAT_MONO16, buffer, buflen*sizeof(short), sampleRate);
free(buffer);
return albuf;
}

static void freeSource (channel* c) {
if (!c) return;
if (c->src && alIsSource(c->src)) {
alSourceStop(c->src);
alDeleteSources(1,&(c->src));
}
c->src=0;
}

static void freeStream (stream* s) {
if (!s || !s->handle || !s->src || !alIsSource(s->src)) return;
SDL_LockMutex(s->mutex);
alSourceStop(s->src);
s->stopped = 1;
SDL_UnlockMutex(s->mutex);
SDL_WaitThread(s->thread,NULL);
SDL_DestroyMutex(s->mutex);
}

static int fillBuffer (ALuint buffer, stream* s) {
if (!buffer || !s || !s->src || !s->handle) return 0;
char* buf = malloc(STREAM_BUFLEN);
int read = s->io.read(s->handle, FORMAT_S16, buf, STREAM_BUFLEN/sizeof(short));
if (read<=0) { 
free(buf); 
if (!s->loop) return 0;
s->io.seek(s->handle, 0, SEEK_SET);
return fillBuffer(buffer, s);
}
alBufferData(buffer, s->channels==2? AL_FORMAT_STEREO16 : AL_FORMAT_MONO16, buf, read*sizeof(short), s->sampleRate);
free(buf);
return 1;
}

static int streamProc (stream* s) {
if (!s || !s->handle) return 1;
int processed, queued;
ALuint buffer;
while(1){
SDL_LockMutex(s->mutex);
if (s->stopped || !s->src || !s->handle) break;
int paused = AL_PLAYING;
alGetSourcei(s->src, AL_SOURCE_STATE, &paused);
SDL_UnlockMutex(s->mutex);
if (paused==AL_INITIAL || paused==AL_PAUSED) { SDL_Delay(1); continue; }
SDL_LockMutex(s->mutex);
queued = 0;
alGetSourcei(s->src, AL_BUFFERS_PROCESSED, &processed);
while (processed>0 || s->bufferIndex<STREAM_NBUFFERS) {
if (processed>0) alSourceUnqueueBuffers(s->src,1,&buffer);
else alGenBuffers(1,&buffer);
if (alGetError()) { SDL_Delay(1); continue; }
if (!fillBuffer(buffer, s)) break;
//alSourcePause(s->src);
alSourceQueueBuffers(s->src,1, &buffer);
processed--;
queued = 1;
s->bufferIndex++;
alGetSourcei(s->src, AL_SOURCE_STATE, &paused);
if (paused!=AL_PLAYING) alSourcePlay(s->src);
}
SDL_UnlockMutex(s->mutex);
if (!queued) SDL_Delay(STREAM_SLEEPTIME);
}
if (s->src && alIsSource(s->src)) {
alSourceStop(s->src);
alDeleteSources(1,&(s->src));
}
s->src = 0;
s->io.close(s->handle, NULL, NULL);
s->handle = 0;
SDL_UnlockMutex(s->mutex);
return 0;
}

static int loadStreamImpl (stream* s, void* handle, audioIO* io, audioInfo* info) {
s->handle = handle;
s->io = *io;
alGenSources(1,&(s->src));
if (!s->src) {
s->io.close(s->handle,NULL, NULL);
return 0;
}
s->type = AL_STREAMING;
s->stopped = 0;
s->loop = 0;
s->bufferIndex = 0;
s->channels = info->channels;
s->sampleRate = info->samplerate;
s->totalLength = info->frames;
s->mutex = SDL_CreateMutex();
s->thread = SDL_CreateThread(streamProc, s);
s->handle = dspAttach(s->handle, &(s->io));
s->handle = s->io.close(s->handle, s->handle, &(s->io));
return 1;
}

device* export deviceOpen (const char* devname) {
device d = { .dev=NULL, .ctx=NULL, .ux=0, .uy=0, .uz=1, .dx=0, .dy=1, .dz=0  };
d.dev = alcOpenDevice(devname);
if (!d.dev) return NULL;
d.ctx = alcCreateContext(d.dev,0);
if (!d.ctx) return 0;
if (!alcMakeContextCurrent(d.ctx)) return 0;
alListener3f(AL_POSITION, 0, 0, 0);
float t[6] = { d.dx, d.dy, d.dz, d.ux, d.uy, d.uz };
alListenerfv(AL_ORIENTATION, t);
curdev = malloc(sizeof(device));
memcpy(curdev, &d, sizeof(device));
return curdev;
}

device* export deviceGetCurrent (void) { 
return curdev;  
}

int export deviceSetCurrent (device* dev) {
if (!alcMakeContextCurrent(dev? dev->ctx : 0)) return 0;
curdev = dev;
return 1;
}

int export deviceClose (device* d) {
if (!d) return;
if (d->ctx) {
if (curdev && curdev->ctx==d->ctx) alcMakeContextCurrent(0);
alcDestroyContext(d->ctx);
d->ctx=NULL;
}
if (d->dev) alcCloseDevice(d->dev);
d->dev=NULL;
free(d);
}

const char* export deviceList (void) {
return alcGetString(NULL, ALC_DEVICE_SPECIFIER);
}

ALuint export sampleLoad (const char* filename) {
audioInfo info = { .format=0 };
audioIO io;
void* handle = audioIOOpenFile(filename, &info, &io);
if (!handle) return NULL;
return loadSampleImpl(handle, &io, &info);
}

void export sampleFree (ALuint albuf) {
if (albuf && alIsBuffer(albuf)) alDeleteBuffers(1,&albuf);
}

double export sampleGetLength (ALuint albuf) {
if (!albuf) return 0;
int sz, sr, b, ch;
alGetBufferi(albuf, AL_FREQUENCY, &sr);
alGetBufferi(albuf, AL_BITS, &b);
alGetBufferi(albuf, AL_CHANNELS, &ch);
alGetBufferi(albuf, AL_SIZE, &sz);
return 8.0 * sz / b / sr / ch;
}

double export sampleGetAttr (ALuint albuf, int attr) {
if (!albuf) return 0;
int i=0;
alGetBufferi(albuf, attr, &i);
return i;
}

channel* export channelNew (ALuint albuf) {
ALuint src = getSourceFromBuffer(albuf);
if (!src) return NULL;
channel* c = malloc(sizeof(channel));
c->src = src;
c->type = AL_STATIC;
c->flags = 0;
return c;
}

/*static int sampleLoopPoints (lua_State* l) {
ALuint albuf = *(ALuint*)luaL_checkudata(l, 1, "sample");
int sz, sr, b, ch;
alGetBufferi(albuf, AL_FREQUENCY, &sr);
alGetBufferi(albuf, AL_BITS, &b);
alGetBufferi(albuf, AL_CHANNELS, &ch);
alGetBufferi(albuf, AL_SIZE, &sz);
double len = 8.0 * sz / b / sr / ch;
if (lua_gettop(l)>=2) {
double start = luaL_checknumber(l,2);
double end = luaL_optnumber(l,3,-0.0);
if (start<0) start+=len;
if (end<0) end+=len;
if (end<=len && start<end) {
int lp[2] = { start*sr, end*sr };
alBufferiv(albuf, AL_LOOP_POINTS_SOFT, lp);
}
lua_settop(l,1);
return 1;
}
else {
int lp[2] = { 0, 0 };
alGetBufferiv(albuf, AL_LOOP_POINTS_SOFT, lp);
double start = 1.0 * lp[0] / sr, end = 1.0 * lp[1] /sr;
lua_settop(l,0);
lua_pushnumber(l,start);
lua_pushnumber(l,end);
return 2;
}}*/

void export channelPlay (channel* c) {
if (!c) return;
alSourcePlay(c->src);
}

void export channelPause (channel* c) {
if (!c) return;
alSourcePause(c->src);
}

stream* export streamLoad (const char* filename) {
audioInfo info = { .format=0 };
audioIO io;
void* handle = audioIOOpenFile(filename, &info, &io);
if (!handle) return NULL;
stream* s = malloc(sizeof(stream));
if (!loadStreamImpl(s, handle, &io, &info)) {
free(s);
return NULL;
}
return s;
}

void export channelGC (channel* c) {
if (!c) return;
static ALuint num=0, max=0, *buf=NULL;
if (c->type==AL_STATIC && (c->flags&CHF_AUTOFREE)) {
if (num+1>=max) buf = realloc(buf, sizeof(ALuint) * (max=num*3/2+2));
buf[num++] = c->src;
free(c);
}
for (int i=0; i<num; i++) {
ALuint src = buf[i];
if (!alIsSource(src) || !channelIsPlaying((channel*)(&src))) {
alDeleteSources(1,&src);
buf[i--] = buf[num--];
}}}

void export channelFree (channel* c) {
if (!c) return;
if (c->type == AL_STREAMING) freeStream(c);
else freeSource(c);
free(c);
}

void export channelStop (channel* c) {
if (!c) return;
if (c->type == AL_STREAMING || (c->flags&CHF_AUTOFREE)) channelFree(c);
else alSourceStop(c->src);
}

int export channelIsPlaying (channel* c) {
if (!c) return 0;
int n=0;
alGetSourcei(c->src, AL_SOURCE_STATE, &n);
return n==AL_PLAYING || n==AL_PAUSED;
}

void export channelWait (channel* c, int max) {
if (!c) return;
int delay=0;
if (!c->src || !alIsSource(c->src)) return;
while (channelIsPlaying(c) && (max<0 || delay<max)) {
SDL_Delay(20);
delay += 20;
}}

double export channelGetLength (channel* c) {
if (!c) return 0;
if (c->type == AL_STREAMING) {
stream* s = (stream*)c;
return 1.0 * s->totalLength / s->sampleRate;
} else {
ALuint buffer = 0;
alGetSourcei(c->src, AL_BUFFER, &buffer);
if (!buffer) return -1;
int sz, sr, b, ch;
alBufferi(buffer, AL_FREQUENCY, &sr);
alBufferi(buffer, AL_BITS, &b);
alBufferi(buffer, AL_CHANNELS, &ch);
alBufferi(buffer, AL_SIZE, &sz);
return 8.0 * sz / b / sr / ch;
}}

int export channelGetType (channel* c) {
if (!c || !c->src) return 0;
return c->type;
}

/*int export channelGetFlags (channel* c) {

if (!c) return 0;
return c->flags;
}

void export channelSetFlags (channel* c, int flags) {
if (!c) return;
c->flags = flags;
}*/

ALuint export channelGetSample (channel* c) {
if (!c || !c->src || c->type!=AL_STATIC) return 0;
ALuint albuf=0;
alGetSourcei(c->src, AL_BUFFER, &albuf);
return albuf;
}

void export channelSetSample (channel* c, ALuint albuf) {
if (!c || !c->src || c->type!=AL_STATIC) return;
alSourceStop(c->src);
alSourcei(c->src, AL_BUFFER, albuf);
}

double export channelGetAttr (channel* c, int attr) {
if (!c || !c->src) return 0;
if (c->type == AL_STREAMING) {
stream* s = (stream*)c;
if (attr == AL_SEC_OFFSET) {
int pos = s->io.seek(s->handle, 0, SEEK_CUR);
return 1.0 * pos / s->sampleRate;
}
else if (attr == AL_LOOPING) return s->loop;
}//AL_STREAMING
if (attr == AL_PANNING) {
/*float f[15] = {0};
alGetListenerfv(AL_ORIENTATION, f);
alGetSourcefv(c->src, AL_POSITION, &(f[12]));
// Vector product of at and up vector, give r (right vector)
f[6] = f[1]*f[5]-f[2]*f[4];
f[7] = f[2]*f[3]-f[5]*f[0];
f[8] = f[0]*f[4]-f[1]*f[3];
// Matrix reduction
//next
for (int i=0; i<15; i++) printf("%.4g, ", f[i]);
printf("\r\n");*/
return -10;
}
else if (attr=AL_AUTOFREE) return (c->flags&CHF_AUTOFREE)? 1:0;
else if (attr == AL_LOOPING || attr == AL_SOURCE_RELATIVE || attr==AL_DIRECT_CHANNELS_SOFT) {
int i = 0;
alGetSourcei(c->src, attr, &i);
return i;
} else {
float f;
alGetSourcef(c->src, attr, &f);
return f;
}}

void export channelSetAttr (channel* c, int attr, double val) {
if (!c || !c->src) return;
if (c->type == AL_STREAMING) {
stream* s = (stream*)c;
if (attr == AL_SEC_OFFSET) {
SDL_LockMutex(s->mutex);
int state = 0;
alGetSourcei(s->src, AL_SOURCE_STATE, &state);
if (state==AL_PLAYING) alSourcePause(s->src);
s->io.seek(s->handle, (int)(val * s->sampleRate), SEEK_SET);
if (state==AL_PLAYING)  alSourcePlay(s->src);
SDL_UnlockMutex(s->mutex);
return;
}
else if (attr == AL_LOOPING) {
s->loop = val!=0;
return;
}}
if (attr==AL_PANNING) {
double a = sin(acos(val));
float f[12] = {0};
alGetListenerfv(AL_ORIENTATION, f);
// Vector product of at and up vector, give r (right vector)
f[6] = f[1]*f[5]-f[2]*f[4];
f[7] = f[2]*f[3]-f[5]*f[0];
f[8] = f[0]*f[4]-f[1]*f[3];
// new position = (val * right) + (a * at)
f[9] = val * f[6] + a * f[0];
f[10] = val * f[7] + a * f[1];
f[11] = val * f[8] + a * f[2];
alSourcei(c->src, AL_SOURCE_RELATIVE, 1);
alSourcefv(c->src, AL_POSITION, &(f[9]));
}
else if (attr==AL_AUTOFREE) {
if (val<=0) c->flags&=~CHF_AUTOFREE;
else c->flags|=CHF_AUTOFREE;
}
else if (attr == AL_LOOPING || attr == AL_SOURCE_RELATIVE || attr == AL_DIRECT_CHANNELS_SOFT) alSourcei(c->src, attr, val);
else alSourcef(c->src, attr, val);
}

void export channelSetAttrv (channel* c, int attr, const float* v) {
if (!c || !c->src) return;
alSourcefv(c->src, attr, v);
}

void export channelGetAttrv (channel* c, int attr, float* v) {
if (!c || !c->src) return;
alGetSourcefv(c->src, attr, v);
}

double export listenerGetAttr (int attr) {
float f; int i;
switch(attr){
case AL_GAIN :
alGetListenerf(attr, &f);
return f;
case AL_DOPPLER_FACTOR :
case AL_DOPPLER_VELOCITY :
case AL_SPEED_OF_SOUND :
return alGetDouble(attr);
case ALC_FREQUENCY :
case ALC_MONO_SOURCES :
case ALC_STEREO_SOURCES :
case ALC_REFRESH :
alcGetIntegerv(curdev->dev, attr, 1, &i);
return i;
}
return -1;
}

void export listenerSetAttr (int attr, double val) {
switch(attr){
case AL_GAIN :
alListenerf(AL_GAIN, val);
break;
case AL_DOPPLER_FACTOR :
alDopplerFactor(val);
break;
case AL_DOPPLER_VELOCITY :
alDopplerVelocity(val);
break;
case AL_SPEED_OF_SOUND :
alSpeedOfSound(val);
break;
}}

void export listenerSetAttrv (int attr, const float* v) {
if (!v) return;
if (attr == AL_DIRECTION && curdev) {
device* d = curdev;
if (!d) return;
d->dx = v[0];
d->dy = v[1];
d->dz = v[2];
float t[6] = { d->dx, d->dy, d->dz, d->ux, d->uy, d->uz };
alListenerfv(AL_ORIENTATION, t);
}
else if (attr == AL_LISTENER_UP && curdev) {
device* d = curdev;
if (!d) return;
d->ux = v[0];
d->uy = v[1];
d->uz = v[2];
float t[6] = { d->dx, d->dy, d->dz, d->ux, d->uy, d->uz };
alListenerfv(AL_ORIENTATION, t);
}
else alListenerfv(attr, v);
}

void export listenerGetAttrv (int attr, float* v) {
if (!v) return;
if (attr == AL_DIRECTION && curdev) {
v[0] = curdev->dx;
v[1] = curdev->dy;
v[2] = curdev->dz;
}
else if (attr == AL_LISTENER_UP && curdev) {
v[0] = curdev->ux;
v[1] = curdev->uy;
v[2] = curdev->uz;
}
else alGetListenerfv(attr, v);
}

/*static int lSourceSetVector (lua_State* l, int param) {
ALuint src = *(ALuint*)luaL_checkudata(l, 1, "channel");
if (lua_gettop(l)>=3) {
double x = luaL_checknumber(l,2);
double y = luaL_checknumber(l,3);
double z = luaL_optnumber(l,4, 0);
alSource3f(src, param, x, y, z);
lua_settop(l,1);
return 1;
}
else {
float x=0, y=0, z=0;
alGetSource3f(src, param, &x, &y, &z);
lua_settop(l,0);
lua_pushnumber(l,x);
lua_pushnumber(l,y);
lua_pushnumber(l,z);
return 3;
}}
static int channelPosition (lua_State* l) { return lSourceSetVector(l, AL_POSITION); }
static int channelDirection (lua_State* l) { return lSourceSetVector(l, AL_DIRECTION); }
static int channelVelocity (lua_State* l) { return lSourceSetVector(l, AL_VELOCITY); }

static int lListenerSetVector (lua_State* l, int param) {
if (lua_gettop(l)>=3) {
double x = luaL_checknumber(l,2);
double y = luaL_checknumber(l,3);
double z = luaL_optnumber(l,4, 0);
alListener3f(param, x, y, z);
lua_settop(l,1);
return 1;
}
else {
float x=0, y=0, z=0;
alGetListener3f(param, &x, &y, &z);
lua_settop(l,0);
lua_pushnumber(l,x);
lua_pushnumber(l,y);
lua_pushnumber(l,z);
return 3;
}}
static int listenerPosition (lua_State* l) { return lListenerSetVector(l, AL_POSITION); }
static int listenerVelocity (lua_State* l) { return lListenerSetVector(l, AL_VELOCITY); }

static int listenerDirection (lua_State* l, int param) {
device* d = luaL_checkudata(l, 1, "device");
if (lua_gettop(l)>=3) {
d->dx = luaL_checknumber(l,2);
d->dy = luaL_checknumber(l,3);
d->dz = luaL_optnumber(l,4, d->dz);
float t[6] = { d->dx, d->dy, d->dz, d->ux, d->uy, d->uz };
alListenerfv(AL_ORIENTATION, t);
lua_settop(l,1);
return 1;
}
else {
float x=0, y=0, z=0;
lua_settop(l,0);
lua_pushnumber(l,d->dx);
lua_pushnumber(l,d->dy);
lua_pushnumber(l,d->dz);
return 3;
}}

static int listenerUp (lua_State* l, int param) {
device* d = luaL_checkudata(l, 1, "device");
if (lua_gettop(l)>=4) {
d->ux = luaL_checknumber(l,2);
d->uy = luaL_checknumber(l,3);
d->uz = luaL_checknumber(l,4);
float t[6] = { d->dx, d->dy, d->dz, d->ux, d->uy, d->uz };
alListenerfv(AL_ORIENTATION, t);
lua_settop(l,1);
return 1;
}
else {
float x=0, y=0, z=0;
lua_settop(l,0);
lua_pushnumber(l,d->ux);
lua_pushnumber(l,d->uy);
lua_pushnumber(l,d->uz);
return 3;
}}

//##
static int sampleMetaIndex (lua_State* l) {
ALuint albuf = *(ALuint*)luaL_checkudata(l, 1, "sample");
const char* nm = luaL_checkstring(l,2);
if (streq(nm, "length") || streq(nm, "duration")) {
int sz, sr, b, ch;
alGetBufferi(albuf, AL_FREQUENCY, &sr);
alGetBufferi(albuf, AL_BITS, &b);
alGetBufferi(albuf, AL_CHANNELS, &ch);
alGetBufferi(albuf, AL_SIZE, &sz);
double len = 8.0 * sz / b / sr / ch;
lua_pushnumber(l,len);
return 1;
}
else if (streq(nm, "frequency") || streq(nm, "sampleRate")) {
int n=0;
alGetBufferi(albuf, AL_FREQUENCY, &n);
lua_pushinteger(l,n);
return 1;
}
else if (streq(nm, "bits") || streq(nm, "resolution")) {
int n=0;
alGetBufferi(albuf, AL_BITS, &n);
lua_pushinteger(l,n);
return 1;
}
else if (streq(nm, "channels")) {
int n=0;
alGetBufferi(albuf, AL_CHANNELS, &n);
lua_pushinteger(l,n);
return 1;
}
else return luaL_getmetafield(l, 1, nm);
}

static int channelMetaIndex (lua_State* l) {
channel* c = luaL_checkudata(l, 1, "channel");
ALuint src = c->src;
const char* nm = luaL_checkstring(l,2);
if (streq(nm, "playing")) {
int i = 0;
alGetSourcei(src, AL_SOURCE_STATE, &i);
lua_pushboolean(l, i==AL_PLAYING);
return 1;
}
else if (streq(nm, "volume")) {
float f = 1;
alGetSourcef(src, AL_GAIN, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "pitch")) {
float f = 1;
alGetSourcef(src, AL_PITCH, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "offset")) {
if (c->type == AL_STREAMING) {
stream* s = (stream*)c;
int pos = s->io.seek(s->handle, 0, SEEK_CUR);
lua_pushnumber(l, 1.0 * pos / s->sampleRate);
return 1;
} else {
float f = -1;
alGetSourcef(src, AL_SEC_OFFSET, &f);
lua_pushnumber(l,f);
return 1;
}}
else if (streq(nm, "length") || streq(nm, "duration")) {
if (c->type == AL_STREAMING) lua_pushnumber(l, 1.0 * oggLength( ((stream*)c)->handle ) / ((stream*)c)->sampleRate);
else {
ALuint buffer = 0;
double len = -1;
alGetSourcei(src, AL_BUFFER, &buffer);
if (buffer) {
int sz, sr, b, ch;
alBufferi(buffer, AL_FREQUENCY, &sr);
alBufferi(buffer, AL_BITS, &b);
alBufferi(buffer, AL_CHANNELS, &ch);
alBufferi(buffer, AL_SIZE, &sz);
len = 8.0 * sz / b / sr / ch;
}
lua_pushnumber(l,len);
}
return 1;
}
else if (streq(nm, "looped")) {
if (c->type == AL_STREAMING) lua_pushboolean(l, ((stream*)c)->loop);
else {
int i = 0;
alGetSourcei(src, AL_LOOPING, &i);
lua_pushboolean(l,!!i);
return 1;
}}
else if (streq(nm, "relative")) {
int i = 0;
alGetSourcei(src, AL_SOURCE_RELATIVE, &i);
lua_pushboolean(l,!!i);
return 1;
}
else if (streq(nm, "direct")) {
int i = 0;
alGetSourcei(src, AL_DIRECT_CHANNELS_SOFT, &i);
lua_pushboolean(l,!!i);
return 1;
}
else if (streq(nm, "sample")) {
if (c->type == AL_STREAMING) return 0;
ALuint albuf = 0;
alGetSourcei(src, AL_BUFFER, &albuf);
lua_pushudata(l, &albuf, sizeof(albuf), "sample");
return 1;
}
else if (streq(nm, "referenceDistance")) {
float f = -1;
alGetSourcef(src, AL_REFERENCE_DISTANCE, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "maxDistance")) {
float f = -1;
alGetSourcef(src, AL_MAX_DISTANCE, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "rolloff")) {
float f = -1;
alGetSourcef(src, AL_ROLLOFF_FACTOR, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "coneInnerAngle")) {
float f = -1;
alGetSourcef(src, AL_CONE_INNER_ANGLE, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "coneOuterAngle")) {
float f = -1;
alGetSourcef(src, AL_CONE_OUTER_ANGLE, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "coneOuterGain")) {
float f = -1;
alGetSourcef(src, AL_CONE_OUTER_GAIN, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "maxGain")) {
float f = -1;
alGetSourcef(src, AL_MAX_GAIN, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "minGain")) {
float f = -1;
alGetSourcef(src, AL_MIN_GAIN, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "autofree")) {
lua_pushboolean(l, c->type==AL_STREAMING || (c->flags&CHF_AUTOFREE));
return 1;
}
return luaL_getmetafield(l, 1, nm);
}

static int channelMetaNewIndex (lua_State* l) {
channel* c = luaL_checkudata(l, 1, "channel");
ALuint src = c->src;
const char* nm = luaL_checkstring(l,2);
if (streq(nm, "volume")) alSourcef(src, AL_GAIN, luaL_checknumber(l,3));
else if (streq(nm, "pitch")) alSourcef(src, AL_PITCH, luaL_checknumber(l,3));
else if (streq(nm, "offset")) {
if (c->type == AL_STATIC) alSourcei(src, AL_SEC_OFFSET, luaL_checknumber(l,3));
else {
double pos = luaL_checknumber(l,3);
stream* s = (stream*)c;
SDL_LockMutex(s->mutex);
int state = 0;
alGetSourcei(s->src, AL_SOURCE_STATE, &state);
if (state==AL_PLAYING) alSourcePause(s->src);
oggSeek(s->handle, (int)(pos * s->sampleRate));
if (state==AL_PLAYING)  alSourcePlay(s->src);
SDL_UnlockMutex(s->mutex);
}}
else if (streq(nm, "looped")) {
if (c->type == AL_STATIC) alSourcei(src, AL_LOOPING, luaL_checkboolean(l,3));
else ((stream*)c)->loop = luaL_checkboolean(l,3);
}
else if (streq(nm, "relative")) alSourcei(src, AL_SOURCE_RELATIVE, luaL_checkboolean(l,3));
else if (streq(nm, "direct")) alSourcei(src, AL_DIRECT_CHANNELS_SOFT, luaL_checkboolean(l,3));
else if (streq(nm, "sample")) {
if (c->type == AL_STREAMING) return 0;
alSourceStop(src);
alSourcei(src, AL_BUFFER, *(ALuint*)luaL_checkudata(l,3,"sample"));
}
else if (streq(nm, "referenceDistance")) alSourcef(src, AL_REFERENCE_DISTANCE, luaL_checknumber(l,3));
else if (streq(nm, "maxDistance")) alSourcef(src, AL_MAX_DISTANCE, luaL_checknumber(l,3));
else if (streq(nm, "rolloff")) alSourcef(src, AL_ROLLOFF_FACTOR, luaL_checknumber(l,3));
else if (streq(nm, "coneInnerAngle")) alSourcef(src, AL_CONE_INNER_ANGLE, luaL_checknumber(l,3));
else if (streq(nm, "coneOuterAngle")) alSourcef(src, AL_CONE_OUTER_ANGLE, luaL_checknumber(l,3));
else if (streq(nm, "coneOuterGain")) alSourcef(src, AL_CONE_OUTER_GAIN, luaL_checknumber(l,3));
else if (streq(nm, "maxGain")) alSourcef(src, AL_MAX_GAIN, luaL_checknumber(l,3));
else if (streq(nm, "minGain")) alSourcef(src, AL_MIN_GAIN, luaL_checknumber(l,3));
else if (streq(nm, "autofree") && c->type==AL_STATIC) {
if (luaL_checkboolean(l,3)) c->flags |= CHF_AUTOFREE;
else c->flags &=~CHF_AUTOFREE;
}
return 0;
}

static int listenerMetaIndex (lua_State* l) {
const char* nm = luaL_checkstring(l,2);
if (streq(nm, "volume")) {
float f = 1;
alGetListenerf(AL_GAIN, &f);
lua_pushnumber(l,f);
return 1;
}
else if (streq(nm, "dopplerFactor")) { lua_pushnumber(l,alGetDouble(AL_DOPPLER_FACTOR)); return 1; }
else if (streq(nm, "dopplerVelocity")) { lua_pushnumber(l,alGetDouble(AL_DOPPLER_VELOCITY)); return 1; }
else if (streq(nm, "speedOfSound")) { lua_pushnumber(l,alGetDouble(AL_SPEED_OF_SOUND)); return 1; }
else if (streq(nm, "devices")) return deviceList(l);
else {
device* d = luaL_checkudata(l, 1, "device");
if (streq(nm, "deviceName")) {
const char* dn = alcGetString(d->dev, ALC_DEVICE_SPECIFIER);
lua_pushstring(l,dn);
return 1;
}
else if (streq(nm, "frequency") || streq(nm, "sampleRate")) {
int i=0;
alcGetIntegerv(d->dev, ALC_FREQUENCY, 1, &i);
lua_pushinteger(l,i);
return 1;
}
else if (streq(nm, "refreshRate")) {
int i=0;
alcGetIntegerv(d->dev, ALC_REFRESH, 1, &i);
lua_pushinteger(l,i);
return 1;
}
else if (streq(nm, "stereoChannels")) {
int i=0;
alcGetIntegerv(d->dev, ALC_STEREO_SOURCES, 1, &i);
lua_pushinteger(l,i);
return 1;
}
else if (streq(nm, "monoChannels")) {
int i=0;
alcGetIntegerv(d->dev, ALC_MONO_SOURCES, 1, &i);
lua_pushinteger(l,i);
return 1;
}}
return luaL_getmetafield(l, 1, nm);
}
static int listenerMetaNewIndex (lua_State* l) {
const char* nm = luaL_checkstring(l,2);
if (streq(nm, "volume")) alListenerf(AL_GAIN, luaL_checknumber(l,3));
else if (streq(nm, "dopplerFactor")) alDopplerFactor(luaL_checknumber(l,3));
else if (streq(nm, "dopplerVelocity")) alDopplerVelocity(luaL_checknumber(l,3));
else if (streq(nm, "speedOfSound")) alSpeedOfSound(luaL_checknumber(l,3));
return 0;
}

int luaopen_oal  (lua_State* l) {
lua_settop(l,0);
luaL_newclass(l, "device");
lregt(l, "position", listenerPosition);
lregt(l, "direction", listenerDirection);
lregt(l, "velocity", listenerVelocity);
lregt(l, "up", listenerUp);
lregt(l, "open", deviceOpen);
lregt(l, "close", deviceClose);
lregt(l, "getDeviceList", deviceList);
lregt(l, "loadSample", sampleLoad);
lregt(l, "loadStream", streamLoad);
lregt(l, "newChannel", channelNew);
lregt(l, "__index", listenerMetaIndex);
lregt(l, "__newindex", listenerMetaNewIndex);

luaL_newclass(l, "sample");
lregt(l, "loopPoints", sampleLoopPoints);
lregt(l, "play", samplePlay);
lregt(l, "free", sampleFree);
lregt(l, "__gc", sampleFree);
lregt(l, "__eq", sampleCompare);
lregt(l, "__index", sampleMetaIndex);
lua_setfield(l, -2, "sample");

luaL_newclass(l, "channel");
lregt(l, "play", channelPlay);
lregt(l, "pause", channelPause);
lregt(l, "stop", channelStop);
lregt(l, "wait", channelWait);
lregt(l, "free", channelFree);
lregt(l, "position", channelPosition);
lregt(l, "velocity", channelVelocity);
lregt(l, "direction", channelDirection);
lregt(l, "__gc", channelFree);
lregt(l, "__index", channelMetaIndex);
lregt(l, "__newindex", channelMetaNewIndex);
lua_setfield(l, -2, "channel");

lua_pushvalue(l,1);
return 1;
}*/
