#include<lua/lua.h>
#include<lua/lualib.h>
#include<lua/lauxlib.h>
#include<stdio.h>
#include<stdlib.h>
#include "../lua2exe.h"

#define streq(a,b) (0==strcmp(a,b))

#define DEFAULT_BASE_EXECUTABLE "lua.lvm"
#ifdef __WIN32
#define DEFAULT_OUTPUT_FILE "a.exe"
#else
#define DEFAULT_OUTPUT_FILE "a.out"
#endif

#define BOOL int
#define TRUE 1
#define FALSE 0

static char* key = NULL;
static int keylen = 0;
static BOOL quiet = FALSE;

static void setkey (const char* k, int hex) {
if (key) free(key);
keylen = strlen(k)/(hex?2:1);
key = malloc(keylen+4);
if (!hex) memcpy(key, k, keylen+1);
else {
for (int i=0; i<keylen; i++) {
char z[3] = { k[2*i], k[2*i+1], 0 };
sscanf(z, "%02x", key+i);
}
key[keylen]=0;
}
if (!quiet) printf("key set to \"%s\"\r\n", key);
}

static int writeChunk (FILE* fp, lua_State* L, BOOL dbg, int* re1, int* re2) {
EmbeddedChunk c;
int length, offset;
lua_pushboolean(L, !dbg);
lua_call(L, 2, 1);
const char* data = lua_tolstring(L, -1, &length);
if (!data) return -1;
int compressed = lzqWriteData(fp, &offset, data, length, key, keylen);
lua_pop(L,1);
if (compressed<0) return -1;
if (re1) *re1 = length;
if (re2) *re2 = compressed;
return offset;
}

static int printVersion () {
printf("%s", 
"Lua2Exe version 1.0 (april 2012)\r\n"
"Copyright  2012, QuentinC (http://quentinc.net/) \r\n"
"Using luajit 2.0.0.10 as lua VM ( 2005-2011, Mike Pall) (http://luajit.org/) \r\n"
); return 0;
}

int printHelp () {
printVersion();
printf("%s", 
"Usage: lua2exe options scripts... \r\n"
"Scripts: scripts to compile and embed into the lua executable. \r\n"
"Options: zero, one or more of the following options: \r\n"
"-f\tRead the list of files to embed from a file.\r\n"
"-i\tRead the list of files to embed from stdin.\r\n"
"-o [file] \tSet output file. Default to " DEFAULT_OUTPUT_FILE "\r\n"
"-b [file] \tSet base lua executable. Default to " DEFAULT_BASE_EXECUTABLE "\r\n"
"-g \tGenerate debug info, disabled by default.\r\n"
"-k \tSet encryption key.\r\n"
"-x \tSet encryption key in hex representation.\r\n"
"-q\tBe quiet, don't output anything except on error.\r\n"
"-h \tPrint this help \r\n"
"-v \tPrint version information\r\n\r\n"
"First script given is considered as being the main one and is automatically run when the generated executable starts. \r\n"
"Other scripts are compiled and embedded, but are only run when called by require. \r\n"
"An embedded script has always the priority against a script file with the same name beside the executable. \r\n"
); return 0;
}

void preparename (char* s) {
char* c = strchr(s, '=');
if (c) *c=0;
c = strrchr(s,'.');
if (c) *c=0;
while (*(++s)) {
if (*s=='/' || *s=='\\') *s='.';
}}

int main (int argc, char** argv) {
lua_State* L = NULL;
char baseFile[300] = DEFAULT_BASE_EXECUTABLE;
char outputFile[300] = DEFAULT_OUTPUT_FILE;
FILE* inputFile = NULL;
int argi = 0;
BOOL keepDebugInfo = FALSE;

if (argc<2) return printHelp();
while (++argi<argc && argv[argi][0]=='-') {
if (streq(argv[argi], "-o")) strcpy(outputFile, argv[++argi]);
else if (streq(argv[argi], "-b")) strcpy(baseFile, argv[++argi]);
else if (streq(argv[argi], "-f")) inputFile = fopen(argv[++argi], "r");
else if (streq(argv[argi], "-i")) inputFile = (FILE*)stdin;
else if (streq(argv[argi], "-k")) setkey(argv[++argi], 0);
else if (streq(argv[argi], "-x")) setkey(argv[++argi], 1);
else if (streq(argv[argi], "-g")) keepDebugInfo = TRUE;
else if (streq(argv[argi], "-q")) quiet = TRUE;
else if (streq(argv[argi], "-h") || streq(argv[argi], "--help") || streq(argv[argi], "-?")) return printHelp();
else if (streq(argv[argi], "-v")) return printVersion();
else fprintf(stderr, "Warning: unknown option: %s\r\n", argv[argi]);
}
if (!quiet) printVersion();

int n, m, buflen = 32, bufpos = 0;
char* buf  = malloc(buflen);
if (!buf) return 1;

FILE* out = fopen(outputFile, "wb");
FILE* in = fopen(baseFile, "rb");
if (!out) { fprintf(stderr, "Couldn't open output file %s\r\n", outputFile); return 1; }
if (!in) { fprintf(stderr, "Couldn't read base file %s\r\n", baseFile); return 1; }
while ((n=fread(buf, 1, buflen, in))>0) fwrite(buf, 1, n, out);
fclose(in);

L = lua_open();
luaL_openlibs(L);
lua_getglobal(L, "string");
lua_getfield(L, -1, "dump");

int output (char* fn) {
if (
!strncmp(fn, "-k=", 3)
|| !strncmp(fn, "-x=", 3)
) {
setkey(fn+3, *fn=='x');
return 0;
}
char* cfn = strchr(fn, '=');
if (!cfn) cfn=fn;
else cfn++;
lua_pushvalue(L,-1);
if (luaL_loadfile(L, cfn)) {
fprintf(stderr, "%s", lua_tostring(L,-1));
return 1;
}
preparename(fn);
int compressed, uncompressed;
if ((n=writeChunk(out, L, keepDebugInfo, &uncompressed, &compressed))<0) { fprintf(stderr, "Problem when writing to output file %s (#%d,%d)\r\n", outputFile, n, __LINE__); return 1; }
if (!(buf = lzqWriteEntry(buf, &bufpos, &buflen, n, fn, 0))) { fprintf(stderr, "Problem when writing to output file %s (#%d,%d)\r\n", outputFile, n, __LINE__); return 1; }
if (!quiet) printf("%s: %d => %d (%d%%) <%s>\r\n", fn, uncompressed, compressed, 100 - 100 * compressed / uncompressed, cfn);
return 0;
}

for (; argi<argc; argi++) {
if (output(argv[argi])) return 1;
}
if (inputFile) {
char line[512]={0};
while (!feof(inputFile) && !ferror(inputFile) && fgets(line,511,inputFile)) {
if (*line=='#' || line==';' || *line=='\n') continue;
char* c = strchr(line,'\n');
if (c) *c=0;
c=line;
while(*c&&*c<=32) c++;
if (strlen(c)<=0) continue;
if (output(c)) return 1;
}
fclose(inputFile);
}

int compressed = lzqWriteData(out, &n, buf, bufpos, key, keylen);
if (!quiet) printf("<header>: %d => %d (%d%%) <header>\r\n", bufpos, compressed, 100 - 100 * compressed / bufpos);
lzqCloseWrite(out, n);
fflush(out);
fclose(out);
if (!quiet) printf("Successfully generated %s.\r\n", outputFile);
return 0;
}
