#include "global.h"
#include<lua/lua.h>
#include<lua/lualib.h>
#include<lua/lauxlib.h>
#include<stdio.h>
#include<stdlib.h>

#define STATUS_INACTIVE 0
#define STATUS_RUNNING 1
#define STATUS_YIELDED 2
#define STATUS_FINISHED 3
#define STATUS_ERROR 4

static const char* STATUS_NAMES[] = {
"inactive", "running", "yielded", "finished", "error"
};

typedef struct {
lua_State* L;
void *thread, *lock, *yield;
int nargs;
unsigned int status :3;
unsigned int perror :1;
unsigned int refs :4;
} StateHandle;

int lua_getbacktrace (lua_State*) ;

static int threadproc (StateHandle* h) {
if (lua_pcall(h->L, h->nargs, LUA_MULTRET, 1)) {
h->status = STATUS_ERROR;
if (h->perror) fprintf(stderr, "%s", lua_tostring(h->L,-1));
}
else h->status = STATUS_FINISHED;
if (lua_gettop(h->L)>=1 && lua_iscfunction(h->L,1) && lua_tocfunction(h->L,1)==lua_getbacktrace) lua_remove(h->L,1);
h->thread = NULL;
semRelease(h->lock);
if (h->refs<=0) {
if (h->L) lua_close(h->L);
free(h);
}
return 0;
}

static int state_run (lua_State* L) {
StateHandle* h = *(StateHandle**)luaL_checkudata(L, 1, "StateHandle");
if (!h || h->status == STATUS_RUNNING || h->status == STATUS_YIELDED) luaL_argerror(L, 1, "thread is running or yielded");
semAcquire(h->lock);
luaL_checkfunction(L,2);
lua_settop(h->L,0);
lua_pushcclosure(h->L, lua_getbacktrace, 0);
lua_interstate_copy(h->L, L, 2);
int top = lua_gettop(L);
for (int i=3; i<=top; i++) lua_interstate_copy(h->L, L, i);
h->status = STATUS_RUNNING;
h->nargs = top -2;
h->thread = threadNew(threadproc,h);
lua_settop(L,1);
return 1;
}

static int state_runfile (lua_State* L) {
StateHandle* h = *(StateHandle**)luaL_checkudata(L, 1, "StateHandle");
if (!h || h->status == STATUS_RUNNING || h->status == STATUS_YIELDED) luaL_argerror(L, 1, "thread is running or yielded");
semAcquire(h->lock);
lua_settop(h->L,0);
lua_pushcclosure(h->L, lua_getbacktrace, 0);
if (luaL_loadfile(h->L, luaL_checkstring(L,2))) return luaL_error(L, lua_tostring(h->L,-1));
int top = lua_gettop(L);
for (int i=3; i<=top; i++) lua_interstate_copy(h->L, L, i);
h->status = STATUS_RUNNING;
h->nargs = top -2;
h->thread = threadNew(threadproc,h);
lua_settop(L,1);
return 1;
}

static int state_require (lua_State* L) {
StateHandle* h = *(StateHandle**)luaL_checkudata(L, 1, "StateHandle");
if (!h || h->status == STATUS_RUNNING) luaL_argerror(L, 1, "thread is running");
semAcquire(h->lock);
const char* varname = luaL_optstring(L,3,NULL);
lua_settop(h->L,0);
lua_getglobal(h->L, "require");
lua_pushstring(h->L, luaL_checkstring(L,2));
if (lua_pcall(h->L, 1, 1, 0)) return luaL_error(L, lua_tostring(h->L,-1));
if (varname) lua_setglobal(h->L, varname);
lua_settop(h->L,0);
semRelease(h->lock);
return 0;
}

static int state_resume (lua_State* L) {
StateHandle* h = *(StateHandle**)luaL_checkudata(L, 1, "StateHandle");
if (!h || !h->yield || h->status!=STATUS_YIELDED) luaL_argerror(L, 1, "thread isn't yielded");
semAcquire(h->lock);
lua_settop(h->L,0);
for (int i=2, n=lua_gettop(L); i<=n; i++) lua_interstate_copy(h->L, L, i);
h->status = STATUS_RUNNING;
semRelease(h->yield); 
return 0;
}

static int state_yield (lua_State* L) {
lua_getreg(L, ".$StateHandle");
StateHandle* h = lua_topointer(L,-1);
if (!h || L!=h->L) luaL_error(L, "Trying to yield the main thread");
h->status = STATUS_YIELDED;
if (!h->yield) semNew(h->yield,0);
semRelease(h->lock);
semAcquire(h->yield); // wait for resume
return lua_gettop(L); // at this point, we have resumed
}

static int wait2 (lua_State* L, int pushResults) {
StateHandle* h = *(StateHandle**)luaL_checkudata(L, 1, "StateHandle");
int timeout = luaL_optint(L,2,0);
if (timeout<=0) semAcquire(h->lock);
else if (semTryAcquire(h->lock, timeout)) {
lua_pushnil(L);
return 1;
}
if (h->status == STATUS_ERROR) return luaL_error(L, lua_tostring(h->L,-1));
if (pushResults) {
lua_settop(L,0);
for (int i=1, n=lua_gettop(h->L); i<=n; i++) lua_interstate_copy(L, h->L, i);
semRelease(h->lock);
return lua_gettop(L);
} else {
semRelease(h->lock);
lua_pushboolean(L,1);
return 1;
}}

static int state_wait (lua_State* L) {
return wait2(L,0);
}

static int state_get (lua_State* L) {
return wait2(L,1);
}

static int state_index (lua_State* L) {
StateHandle* h = *(StateHandle**)luaL_checkudata(L, 1, "StateHandle");
if (!h || !h->L) return 0;
if (lua_type(L,2)==LUA_TNUMBER) {
int n = luaL_checkint(L,2);
if (n<1) luaL_argerror(L, 2, "invalid index");
if (h->status == STATUS_RUNNING) {
lua_pushinteger(L,-1);
lua_insert(L,2);
wait2(L,0);
}
semAcquire(h->lock);
lua_interstate_copy(L, h->L, n);
semRelease(h->lock);
return 1;
}
const char* nm = lua_tostring(L,2);
if (luaL_getmetafield(L,1,nm)) return 1;
else if (streq(nm, "running")) lua_pushboolean(L, h->status == STATUS_RUNNING);
else if (streq(nm, "yielded")) lua_pushboolean(L, h->status == STATUS_YIELDED);
else if (streq(nm, "finished")) lua_pushboolean(L, h->status == STATUS_FINISHED);
else if (streq(nm, "status")) lua_pushstring(L, STATUS_NAMES[h->status]);
else if (streq(nm, "printErrors")) lua_pushboolean(L, h->perror);
else if (streq(nm, "error") && h->status == STATUS_ERROR) {
semAcquire(h->lock);
lua_interstate_copy(L, h->L, lua_gettop(h->L));
semRelease(h->lock);
return 1;
}
else lua_pushnil(L);
return 1;
}

static int state_newindex (lua_State* L) {
StateHandle* h = *(StateHandle**)luaL_checkudata(L, 1, "StateHandle");
if (!h || !h->L) return 0;
const char* nm = luaL_checkstring(L,2);
if (streq(nm, "printErrors")) h->perror = luaL_checkboolean(L,3);
return 0;
}

static int state_gc (lua_State* L) {
StateHandle* h = *(StateHandle**)luaL_checkudata(L, 1, "StateHandle");
if (!h || --h->refs>0 || h->status == STATUS_RUNNING) return 0;
if (h->L) lua_close(h->L);
free(h);
}

static int state_create (lua_State* L) {
StateHandle* h = malloc(sizeof(StateHandle));
h->L = lua_open2(NULL, L);
lua_pushlightuserdata(h->L, h);
lua_setreg(h->L, ".$StateHandle");
h->status = STATUS_INACTIVE;
h->thread = NULL;
h->nargs = -1;
h->perror = 1;
h->refs = 1;
h->yield = NULL;
semNew(h->lock,1);
lua_pushudata(L, &h, sizeof(StateHandle*), "StateHandle");
return 1;
}

int export luaopen_thread (lua_State* L) {
lua_getglobal(L,"string");
lua_getfield(L, -1, "dump");
lua_setreg(L,"string.dump");
lua_pop(L,1);
luaL_newclass(L, "StateHandle");
regt(L, "get", state_get);
regt(L, "run", state_run);
regt(L, "runfile", state_runfile);
regt(L, "require", state_require);
regt(L, "resume", state_resume);
regt(L, "wait", state_wait);
regt(L, "new", state_create);
regt(L, "yield", state_yield);
regt(L, "free", state_gc);
regt(L, "__index", state_index);
regt(L, "__newindex", state_newindex);
regt(L, "__gc", state_gc);
return 1;
}
